@@ -3,7 +3,7 @@ from typing import Any, Iterable, Union, List  # To make stuff look nicer
 from types import MethodType  # Used to attach functions to the class
 
 # Custom Classes:
-from .pymadClasses import madObject, madFunctor, madReference
+from .pymadClasses import madObject, madFunction, madReference
 from .mad_process import mad_process
 
 # TODO: Make it so that MAD does the loop for variables not python (speed)
@@ -16,7 +16,7 @@ class MAD(object):  # Review private and public
         py_name: A string indicating the name of the reference to python from MAD-NG, for communication from MAD-NG to Python.
     """
 
-    def __init__(self, py_name: str = "py", mad_path: str = None, debug: bool = False):
+    def __init__(self, py_name: str = "py", mad_path: str = None, debug: bool = False, ipython_use_jedi: bool = False):
         """Create a MAD Object to interface with MAD-NG.
 
         The modules MADX, elements, sequence, mtable, twiss, beta0, beam, survey, object, track, match are imported into
@@ -29,10 +29,21 @@ class MAD(object):  # Review private and public
                 (default = None)
             debug (bool): Sets debug mode on or off
                 (default = False)
+            ipython_use_jedi (bool): Allow ipython to use jedi in tab completion, will be slower and may result in MAD-NG throwing errors
+                (default = False)
 
         Returns:
             A MAD object, allowing for communication with MAD-NG
         """
+        #Stop jedi running getattr on my classes...
+        self.ipython_use_jedi = ipython_use_jedi
+        if not ipython_use_jedi: 
+            try:
+                shell = get_ipython().__class__.__name__
+                if shell == 'TerminalInteractiveShell':
+                    get_ipython().Completer.use_jedi = False
+            except NameError:
+                pass
         self.__process = mad_process(py_name, mad_path, debug, self)
         self.send(
             """
@@ -49,7 +60,7 @@ class MAD(object):  # Review private and public
         modulesToImport = [
             "MAD",  # Need MAD.MAD?
             "MADX",
-            "elements",
+            "element",
             "sequence",
             "mtable",
             "twiss",
@@ -61,7 +72,6 @@ class MAD(object):  # Review private and public
             "match",
         ]
         self.load("MAD", modulesToImport)
-        self.load("MAD.element")
 
     # ------------------------------------------------------------------------------------------#
 
@@ -159,6 +169,17 @@ class MAD(object):  # Review private and public
         """
         self.__process.send_ctpsa(monos, coefficients)
 
+    def __errhdlr(self, on_off: bool):
+        self.send(f"py:__err("+ str(on_off).lower() + ")")
+
+    def __safe_send_recv(func):
+        def safe_send_recv(self, *args):
+            self.__errhdlr(True)
+            res = func(self, *args)
+            self.__errhdlr(False)
+            return res
+        return safe_send_recv
+
     def load(self, module: str, vars: List[str] = []):
         """Import modules into the MAD-NG environment
 
@@ -172,38 +193,38 @@ class MAD(object):  # Review private and public
         """
         script = ""
         if vars == []:
-            vars = [x.strip("(...)") for x in dir(madReference(module, self))]
+            vars = [x.strip("()") for x in dir(madReference(module, self))]
+        elif isinstance(vars, str):
+            vars = [vars]
         for className in vars:
             script += f"""{className} = {module}.{className}\n"""
         self.__process.send(script)
 
     def __getattr__(self, item):
         if item[0] == "_" and not item == "__last__":
-            raise (AttributeError(item))
+            raise AttributeError(item)
         return self.recv_vars(item)
 
     # -----------------------------Make the class work like a dictionary----------------------------#
     def __setitem__(self, varName: str, var: Any) -> None:
         if isinstance(varName, tuple):
-            nameList = list(varName)
+            varName = list(varName)
             if isinstance(var, madReference):
-                varList = [
+                var = [
                     type(var)(var.__name__ + f"[{i+1}]", self)
-                    for i in range(len(nameList))
+                    for i in range(len(varName))
                 ]
             else:
-                varList = list(var)
-            if len(varList) != len(nameList):
+                var = list(var)
+            if len(var) != len(varName):
                 raise ValueError(
                     "Incorrect number of values to unpack, received",
-                    len(varList),
+                    len(var),
                     "variables and",
-                    len(nameList),
+                    len(varName),
                     "keys",
                 )
-            self.send_vars(nameList, varList)
-        else:
-            self.send_vars(varName, var)
+        self.send_vars(varName, var)
 
     def __getitem__(self, varName: str) -> Any:
         return self.recv_vars(varName)
@@ -287,7 +308,7 @@ class MAD(object):  # Review private and public
         returnVars = []
         for i in range(len(names)):
             if names[i][:2] != "__" or "__last__" in names[i]:  # Check for private variables
-                self.__process.send(f"{self.py_name}:send({names[i]})")
+                self.__process.send(f"{self.py_name}:__err(true):send({names[i]}):__err(false)")
                 returnVars.append(self.__process.recv(names[i]))
         
         if lst_rtn:
@@ -314,7 +335,7 @@ class MAD(object):  # Review private and public
         self.__process.send(
             f"__last__ = __mklast__({self.__to_MAD_string(funcName)}({self.__getArgsAsString(*args)}))\n"
         )
-        return madObject("__last__", self)
+        return madReference("__last__", self)
 
     def MAD_lambda(self, arguments: List[str], expression: str):
         """Create a small anonymous function in MAD-NG named ``__last__``.
@@ -335,7 +356,7 @@ class MAD(object):  # Review private and public
                 result += self.__to_MAD_string(arg) + ","
             result = result[:-1]
         self.send(result + " -> " + expression)
-        return madFunctor("__last__", self)
+        return madFunction("__last__", self)
 
     # -------------------------------------------------------------------------------------------------------------#
 
@@ -419,6 +440,7 @@ class MAD(object):  # Review private and public
     def __dir__(self) -> Iterable[str]:
         pyObjs = [x for x in super(MAD, self).__dir__() if x[0] != "_"]
         pyObjs.extend(self.env())
+        pyObjs.extend(dir(self.recv_vars("_G")))
         return pyObjs
 
     def env(self) -> List[str]:
